/*
    $Id: callbacks.c,v 1.1 2000/03/09 08:38:01 root Exp root $

    gtsl, temperature sensor data logger
    Copyright (C) 2000  James Cameron (quozl@us.netrek.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    http://quozl.us.netrek.org/ts/
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#define MAXSENSOR 4

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

FILE *input;
int fd;
guint in;
guint dog;
guint ndm;

guint context;
GtkStatusbar *statusbar;
GtkHScale *hscale[MAXSENSOR];
char *device = NULL;

/* code theft from gfocustimer follows */
gboolean
on_message_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET(widget)));
  return FALSE;
}


void
on_message_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET(button)));
}


GtkWidget*
create_message ()
{
  GtkWidget *message;
  GtkWidget *dialog_vbox1;
  GtkWidget *message_label;
  GtkWidget *dialog_action_area1;
  GtkWidget *hbuttonbox2;
  GtkWidget *message_ok;

  message = gtk_dialog_new ();
  gtk_widget_set_name (message, "message");
  gtk_object_set_data (GTK_OBJECT (message), "message", message);
  GTK_WINDOW (message)->type = GTK_WINDOW_DIALOG;
  gtk_window_set_title (GTK_WINDOW (message), "unset title");
  gtk_window_set_position (GTK_WINDOW (message), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (message), TRUE, TRUE, TRUE);
  gtk_window_set_wmclass (GTK_WINDOW (message), "gfocustimermessage", "gfocustimermessage");

  dialog_vbox1 = GTK_DIALOG (message)->vbox;
  gtk_widget_set_name (dialog_vbox1, "dialog_vbox1");
  gtk_object_set_data (GTK_OBJECT (message), "dialog_vbox1", dialog_vbox1);
  gtk_widget_show (dialog_vbox1);

  message_label = gtk_label_new ("unset label");
  gtk_widget_set_name (message_label, "message_label");
  gtk_object_set_data (GTK_OBJECT (message), "message_label", message_label);
  gtk_widget_show (message_label);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), message_label, FALSE, FALSE, 0);
  gtk_misc_set_padding (GTK_MISC (message_label), 8, 8);

  dialog_action_area1 = GTK_DIALOG (message)->action_area;
  gtk_widget_set_name (dialog_action_area1, "dialog_action_area1");
  gtk_object_set_data (GTK_OBJECT (message), "dialog_action_area1", dialog_action_area1);
  gtk_widget_show (dialog_action_area1);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);

  hbuttonbox2 = gtk_hbutton_box_new ();
  gtk_widget_set_name (hbuttonbox2, "hbuttonbox2");
  gtk_object_set_data (GTK_OBJECT (message), "hbuttonbox2", hbuttonbox2);
  gtk_widget_show (hbuttonbox2);
  gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbuttonbox2, TRUE, TRUE, 0);

  message_ok = gtk_button_new_with_label ("OK");
  gtk_widget_set_name (message_ok, "message_ok");
  gtk_object_set_data (GTK_OBJECT (message), "message_ok", message_ok);
  gtk_widget_show (message_ok);
  gtk_container_add (GTK_CONTAINER (hbuttonbox2), message_ok);
  GTK_WIDGET_SET_FLAGS (message_ok, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (message_ok);
  gtk_widget_grab_default (message_ok);

  gtk_signal_connect (GTK_OBJECT (message), "delete_event",
                      GTK_SIGNAL_FUNC (on_message_delete_event),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (message_ok), "clicked",
                      GTK_SIGNAL_FUNC (on_message_ok_clicked),
                      NULL);

  return message;
}


/* display a message window, which self deletes when user hits ok */
void message (char *title, char *text)
{
  GtkWidget *widget, *label;

  widget = create_message ();
  gtk_window_set_title (GTK_WINDOW (widget), title);
  label = lookup_widget (widget, "message_label");
  gtk_label_set (GTK_LABEL(label), text);
  gtk_widget_show (widget);
}


/* gtsl new code begins here */


/* accept the device argument from main() */
void gtsl_set_device(char *argv)
{
  device = argv;
}


/* purge a specific message from the status bar stack */
static int purge(gpointer data)
{
  gtk_statusbar_remove (statusbar, context, (guint) data);
  return FALSE;
}


#define TRYSETSENSITIVE
#define TRYHIDING

static void silence(int number)
{
#ifdef TRYSETSENSITIVE
  gtk_widget_set_sensitive(GTK_WIDGET(hscale[number]), FALSE);
#endif
#ifdef TRYHIDING
  gtk_widget_hide(GTK_WIDGET(hscale[number]));
#endif
}

static int silence_timer(gpointer data)
{
  silence((int) data);
  return FALSE;
}

static void silent()
{
  silence(0);
  silence(1);
  silence(2);
  silence(3);
}


/* no data received within watchdog interval, publish it */
static int nodata(gpointer data)
{
  ndm = gtk_statusbar_push (statusbar, context, 
			    "Warning: no data being received");
  silent();
  return FALSE;
}


/* called by gdk/gtk event handler when our fd becomes readable */
static void reader (gpointer data, gint source, GdkInputCondition condition)
{
  int number;
  time_t now;
  struct tm *tm;
  char *p, buffer[128];
  float value;
  static int timeouts[MAXSENSOR] = {0, 0, 0, 0};

  /* get a line of data from the sensor */
  p = fgets(buffer, 128, input);
  if (p == NULL) {
    char *text;

    text = g_strdup_printf("fgets: %s\n\n"
			   "Indicates a problem reading from the port.\n"
			   "The program will no longer display data.\n",
			   strerror(errno));
    message("gtsl: fgets failure", text);
    g_free(text);
    gtk_input_remove(in);
    fclose(input);
    return;
  }

  /* remove the terminating lf or cr */
  p = buffer + strlen(buffer) - 1;
  while (*p == '\r' || *p == '\n') {
    *p = '\0';
    p--;
    if (p < buffer) break;
  }
  
  /* remove preceeding lf or cr */
  p = buffer;
  while (*p == '\r' || *p == '\n') p++;
  
  /* establish timestamp for this sample */
  now = time(NULL);
  tm = localtime(&now);
  
  /* check packet header for reset detection */
  if (*p == 'R') {
    char *message = g_strdup_printf("Microcode %s", p+2);
    gtk_timeout_add(4000, purge, 
		    (gpointer) gtk_statusbar_push(statusbar, context, 
						  message));
    g_free(message);
  }
  
  /* ignore any other packet header other than sensor reports */
  if (!isdigit(*p)) return;
  
  /* decode the sensor number */
  number = atoi(p)-1;
  if (number < 0 || number > MAXSENSOR-1) return;
  
  /* check for missing blank, ignore the line as corrupt */
  p++;
  if (*p != ' ') return;
  
  /* check for verbose mode additional data, and skip it */
  if (strlen(p) > 9) p += 11;

  /* read and decode the data */
  p++;
  sscanf(p, "%f", &value);

  /* display it */
  {
    GtkAdjustment *adjustment;
#ifdef TRYSETSENSITIVE
    gtk_widget_set_sensitive(GTK_WIDGET(hscale[number]), TRUE);
#endif
#ifdef TRYHIDING
    gtk_widget_show(GTK_WIDGET(hscale[number]));
#endif
    adjustment = gtk_range_get_adjustment(GTK_RANGE(hscale[number]));
    gtk_adjustment_set_value (adjustment, value);
    gtk_range_set_adjustment(GTK_RANGE(hscale[number]), adjustment);
  }

  /* reschedule per sensor timeout */
  if (timeouts[number] != 0) gtk_timeout_remove(timeouts[number]);
  timeouts[number] = gtk_timeout_add(5000, silence_timer, (gpointer) number);

  /* now that we have data, remove "no data" status message */
  if (ndm != 0) {
    gtk_statusbar_remove (statusbar, context, ndm);
    ndm = 0;
  }

  /* reschedule no data watchdog */
  gtk_timeout_remove(dog);
  dog = gtk_timeout_add(1500, nodata, NULL);
}


/* start listening */
static void initialise ()
{
  struct termios termios;

  /* allow device to default */
  if (device == NULL) device = "/dev/ttyS0";

  /* open the serial port, causing DTR to be raised and the sensor to run */
  input = fopen(device, "r" );
  if (input == NULL) {
    char *text;

    text = g_strdup_printf("fopen: %s: %s\n\n"

			   "Could not open the %s tty device to accept data\n"
			   "from the microcontroller.  You may need to use\n"
			   "the command line option to specify a device.\n\n"
			   "For example; gtsl -f /dev/ttyS2\n", device,
			   strerror(errno), device);
    message("gtsl: fopen failure", text);
    g_free(text);
    return;
  }

  /* get and set the serial port characteristics */
  fd = fileno(input);
  if (tcgetattr(fd, &termios) < 0) {
    char *text;

    text = g_strdup_printf("tcgetattr: %s: %s\n\n"

			   "Could not obtain tty attributes of %s.\n\n"
			   "This likely means the device you asked me to\n"
			   "use is not a tty, or the default ttyS0 is not\n"
			   "a valid serial port on your machine.\n\n"
			   "You need to provide a different device on the\n"
			   "command line as an option; for example\n\n"
			   "# gtsl -f /dev/ttyS1\n\n",
			   device, strerror(errno), device);

    message("gtsl: tcgetattr failure", text);
    g_free(text);
    return;
  }

  if (cfsetospeed(&termios, B2400) < 0) {
    perror("cfsetospeed");
    exit(1);
  }

  if (cfsetispeed(&termios, B2400) < 0) {
    perror("cfsetispeed");
    exit(1);
  }

  if (tcsetattr(fd, TCSANOW, &termios) < 0) {
    perror("tcsetattr");
    exit(1);
  }

  /* inform gtk that we want to be told when data arrives on fd */
  in = gtk_input_add_full(fd, GDK_INPUT_READ, reader, NULL, NULL, NULL);
}


static int restart(gpointer data)
{
  initialise();
  return FALSE;
}


void
on_reset_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  fclose(input);
  silent();
  gtk_input_remove(in);
  gtk_timeout_remove(dog);
  gtk_timeout_add(2000, restart, NULL);
  gtk_timeout_add(3000, purge,
		  (gpointer) gtk_statusbar_push (statusbar, 
						 context, 
						 "Resetting circuit ..."));
}


void
on_exit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_main_quit();
}


void
on_gtsl_map                            (GtkWidget       *widget,
                                        gpointer         user_data)
{
  statusbar = GTK_STATUSBAR(lookup_widget(widget, "statusbar"));
  context = gtk_statusbar_get_context_id (statusbar, "reset");

  hscale[0] = GTK_HSCALE(lookup_widget(widget, "hscale1"));
  hscale[1] = GTK_HSCALE(lookup_widget(widget, "hscale2"));
  hscale[2] = GTK_HSCALE(lookup_widget(widget, "hscale3"));
  hscale[3] = GTK_HSCALE(lookup_widget(widget, "hscale4"));
  silent();
  initialise();
  gtk_statusbar_push (statusbar, context, "Running");
  dog = gtk_timeout_add(3000, nodata, NULL);
  ndm = 0;
}


gboolean
on_gtsl_delete_event                   (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_main_quit();
  return FALSE;
}


void
on_about_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  message ("About gtsl",
"gtsl, graphical temperature sensor logger, version 0.1,\n"
"Copyright (C) 2000 James Cameron, <quozl@us.netrek.org>\n"
"\n"
"gtsl comes with ABSOLUTELY NO WARRANTY; for details\n"
"see source.  This is free software, and you are welcome to \n"
"redistribute it under certain conditions; see source for details.\n"
"\n"
"This program is designed to operate specifically with a\n"
"preprogrammed microcontroller.\n\n"
"See http://quozl.us.netrek.org/ts/ for more information.\n"
"\n"
"Problem reports, suggestions, and patches are welcome.\n");
}

